/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.components;

import java.util.List;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;


/**
 * @author Guillaume Porcher
 */
public abstract class JoinConditionComponent {

  private static class OrJoinConditionComponent extends JoinConditionComponent {
    private JoinConditionComponent c1;
    private JoinConditionComponent c2;

    public OrJoinConditionComponent(final JoinConditionComponent c1, final JoinConditionComponent c2) {
      this.c1 = c1;
      this.c2 = c2;
    }

    @Override
    public String getExpression() {
      return "(" + this.c1.getExpression() + " or " + this.c2.getExpression() + ")";
    }

    @Override
    protected JoinConditionComponent removeLinkComponent(final LinkComponent linkComponent) {
      final JoinConditionComponent newC1 = this.c1.removeLinkComponent(linkComponent);
      final JoinConditionComponent newC2 = this.c2.removeLinkComponent(linkComponent);
      if (newC1 == null || newC2 == null) {
        return null;
      }
      return new OrJoinConditionComponent(newC1, newC2);
    }

    @Override
    protected boolean replaceLinkComponent(final LinkComponent linkComponent, final JoinConditionComponent joinCondition) {
      if ((this.c1 instanceof LinkJoinConditionComponent)
              && ((LinkJoinConditionComponent) this.c1).l.equals(linkComponent)) {
        this.c1 = joinCondition;
        return true;
      }
      if ((this.c2 instanceof LinkJoinConditionComponent)
              && ((LinkJoinConditionComponent) this.c2).l.equals(linkComponent)) {
        this.c2 = joinCondition;
        return true;
      }
      return this.c1.replaceLinkComponent(linkComponent, joinCondition)
              || this.c2.replaceLinkComponent(linkComponent, joinCondition);
    }
  }

  private static class AndJoinConditionComponent extends JoinConditionComponent {
    private JoinConditionComponent c1;
    private JoinConditionComponent c2;

    public AndJoinConditionComponent(final JoinConditionComponent c1, final JoinConditionComponent c2) {
      this.c1 = c1;
      this.c2 = c2;
    }

    @Override
    public String getExpression() {
      return "(" + this.c1.getExpression() + " and " + this.c2.getExpression() + ")";
    }

    @Override
    protected JoinConditionComponent removeLinkComponent(final LinkComponent linkComponent) {
      final JoinConditionComponent newC1 = this.c1.removeLinkComponent(linkComponent);
      final JoinConditionComponent newC2 = this.c2.removeLinkComponent(linkComponent);
      if (newC1 == null) {
        return newC2;
      }
      if (newC2 == null) {
        return newC1;
      }
      return new AndJoinConditionComponent(newC1, newC2);
    }

    @Override
    protected boolean replaceLinkComponent(final LinkComponent linkComponent, final JoinConditionComponent joinCondition) {
      if ((this.c1 instanceof LinkJoinConditionComponent)
              && ((LinkJoinConditionComponent) this.c1).l.equals(linkComponent)) {
        this.c1 = joinCondition;
        return true;
      }
      if ((this.c2 instanceof LinkJoinConditionComponent)
              && ((LinkJoinConditionComponent) this.c2).l.equals(linkComponent)) {
        this.c2 = joinCondition;
        return true;
      }
      return this.c1.replaceLinkComponent(linkComponent, joinCondition)
              || this.c2.replaceLinkComponent(linkComponent, joinCondition);
    }
  }

  private static class LinkJoinConditionComponent extends JoinConditionComponent {
    private final LinkComponent l;

    public LinkJoinConditionComponent(final LinkComponent l) {
      this.l = l;
    }

    @Override
    public String getExpression() {
      return "$" + this.l.getName();
    }

    @Override
    protected JoinConditionComponent removeLinkComponent(final LinkComponent linkComponent) {
      if (linkComponent == this.l) {
        return null;
      }
      return new LinkJoinConditionComponent(this.l);
    }

    @Override
    protected boolean replaceLinkComponent(final LinkComponent linkComponent, final JoinConditionComponent joinCondition) {
      return false;
    }
  }

  public abstract String getExpression();

  protected abstract JoinConditionComponent removeLinkComponent(LinkComponent linkComponent);

  protected abstract boolean replaceLinkComponent(LinkComponent linkComponent, JoinConditionComponent joinCondition);

  /**
   * @param createdLinkComponents
   */
  public JoinConditionComponent addOrLinks(final List<LinkComponent> links) {
    JoinConditionComponent j = this;
    for (final LinkComponent link : links) {
      j = new OrJoinConditionComponent(new LinkJoinConditionComponent(link), j);
    }
    return j;
  }

  public static void addNewAndLink(final BpmnMap bpmnMap, final LinkComponent linkComponent) {
    JoinConditionComponent.addNewAndCondition(
            bpmnMap, new LinkJoinConditionComponent(linkComponent), linkComponent.getTargetNode());
  }

  public static void addNewAndCondition(
          final BpmnMap bpmnMap, final JoinConditionComponent joinConditionComponent, final TFlowNode targetNode) {
    final JoinConditionComponent j;
    final JoinConditionComponent previousJ = bpmnMap.getJoinCondition(targetNode);
    if (previousJ != null) {
      j = new AndJoinConditionComponent(previousJ, joinConditionComponent);
    } else {
      j = joinConditionComponent;
    }
    bpmnMap.getJoinComponents().put(targetNode, j);
  }

  public static void addNewOrLink(final BpmnMap bpmnMap, final LinkComponent linkComponent) {
    JoinConditionComponent.addNewOrCondition(
            bpmnMap, new LinkJoinConditionComponent(linkComponent), linkComponent.getTargetNode());
  }

  public static void addNewOrCondition(
          final BpmnMap bpmnMap, final JoinConditionComponent joinConditionComponent, final TFlowNode targetNode) {
    final JoinConditionComponent j;
    final JoinConditionComponent previousJ = bpmnMap.getJoinCondition(targetNode);
    if (previousJ != null) {
      j = new OrJoinConditionComponent(previousJ, joinConditionComponent);
    } else {
      j = joinConditionComponent;
    }
    bpmnMap.getJoinComponents().put(targetNode, j);
  }

  public static JoinConditionComponent getOrCondition(final List<LinkComponent> linkComponents) {
    JoinConditionComponent j = null;
    for (final LinkComponent l : linkComponents) {
      if (j == null) {
        j = new LinkJoinConditionComponent(l);
      } else {
        j = new OrJoinConditionComponent(j, new LinkJoinConditionComponent(l));
      }
    }
    return j;
  }

  public static void removeLink(final BpmnMap bpmnMap, final LinkComponent linkComponent) {
    final JoinConditionComponent j = bpmnMap.getJoinCondition(linkComponent.getTargetNode());
    if (j != null) {
      final JoinConditionComponent newJ = j.removeLinkComponent(linkComponent);
      if (newJ != null) {
        bpmnMap.getJoinComponents().put(linkComponent.getTargetNode(), newJ);
      } else {
        bpmnMap.getJoinComponents().remove(linkComponent.getTargetNode());
      }
    }
  }

  /**
   * @param bpmnMap
   * @param linkComponent
   * @param joinCondition
   * @param targetNode
   */
  public static void replaceCondition(
          final BpmnMap bpmnMap,
          final LinkComponent linkComponent,
          final JoinConditionComponent joinCondition,
          final TFlowNode targetNode) {
    final JoinConditionComponent j = bpmnMap.getJoinCondition(linkComponent.getTargetNode());
    if (j == null) {
      JoinConditionComponent.addNewAndCondition(bpmnMap, joinCondition, targetNode);
    } else if ((j instanceof LinkJoinConditionComponent) && ((LinkJoinConditionComponent) j).l.equals(linkComponent)) {
      bpmnMap.getJoinComponents().put(targetNode, joinCondition);
    } else if (!j.replaceLinkComponent(linkComponent, joinCondition)) {
      JoinConditionComponent.addNewAndCondition(bpmnMap, joinCondition, targetNode);
    }
  }


}
