/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.generators;

import java.util.Collection;
import java.util.List;

import javax.xml.namespace.QName;

import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpel.Assign;
import org.ow2.orchestra.jaxb.bpel.Documentation;
import org.ow2.orchestra.jaxb.bpel.FromParts;
import org.ow2.orchestra.jaxb.bpel.Invoke;
import org.ow2.orchestra.jaxb.bpel.Scope;
import org.ow2.orchestra.jaxb.bpel.Sequence;
import org.ow2.orchestra.jaxb.bpel.TActivity;
import org.ow2.orchestra.jaxb.bpel.ToParts;
import org.ow2.orchestra.jaxb.bpel.Variable;
import org.ow2.orchestra.jaxb.bpel.Variables;
import org.ow2.orchestra.jaxb.bpmn.DataInputAssociation;
import org.ow2.orchestra.jaxb.bpmn.DataOutputAssociation;
import org.ow2.orchestra.jaxb.bpmn.MessageFlow;
import org.ow2.orchestra.jaxb.bpmn.TServiceTask;


/**
 * @author Guillaume Porcher
 */
class InvokeGenerator {
  private final BpelGenerator bpelGenerator;

  InvokeGenerator() {
    this.bpelGenerator = new BpelGenerator();
  }

  /**
   * @param bpelGenerator
   */
  InvokeGenerator(final BpelGenerator bpelGenerator) {
    this.bpelGenerator = bpelGenerator;
  }

  TActivity createInvokeActivity(final BpmnMap bpmnMap, final TServiceTask serviceTask, final QName processQName) {
    final Invoke invoke = new Invoke();
    final Assign inAssign = new Assign();
    inAssign.setName(serviceTask.getId() + "_dataInput");
    final Documentation inAssignDoc = new Documentation();
    inAssignDoc.getContent().add("DataInput assignment for " + serviceTask.getName());
    inAssign.getDocumentations().add(inAssignDoc);
    final Assign outAssign = new Assign();
    outAssign.setName(serviceTask.getId() + "_dataOutput");
    final Documentation outAssignDoc = new Documentation();
    outAssignDoc.getContent().add("DataOutput assignment for " + serviceTask.getName());
    outAssign.getDocumentations().add(outAssignDoc);
    final Scope scope = new Scope();
    scope.setName(serviceTask.getId() + "_scope");
    final Documentation scopeDoc = new Documentation();
    scopeDoc.getContent().add("Scope for " + serviceTask.getName());
    scope.getDocumentations().add(scopeDoc);

    final Variables variables = new Variables();
    scope.setVariables(variables);

    invoke.setName(serviceTask.getId());
    final Documentation doc = new Documentation();
    doc.getContent().add(serviceTask.getName());
    invoke.getDocumentations().add(doc);

    if (serviceTask.getOperationRef() == null) {
      throw new TransformerException("ServiceTasks must define an operation.");
    }
    final QName operationQName = serviceTask.getOperationRef();
    final Collection<MessageFlow> messageFlows = bpmnMap.getOutgoingMessageFlows(serviceTask, processQName);
    if (messageFlows.size() != 1) {
      throw new TransformerException("Only one outgoing message flow supported for serviceTask " + serviceTask.getId());
    }
    final MessageFlow messageFlow = messageFlows.iterator().next();
    final QName participantQName = messageFlow.getTargetRef();
    final OperationReference invokeOperationReference =
      this.bpelGenerator.getOperationReference(bpmnMap, operationQName, participantQName);

    if (invokeOperationReference == null) {
      throw new TransformerException("operation " + operationQName + " not found");
    }
    invoke.setOperation(invokeOperationReference.getOperation().getName());
    invoke.setPartnerLink(invokeOperationReference.getPartnerLinkName());
    invoke.setPortType(invokeOperationReference.getPortTypeQName());

    // set fromParts
    final List<DataOutputAssociation> dataOutputAssociations = serviceTask.getDataOutputAssociations();
    if (!dataOutputAssociations.isEmpty()) {
      final Object dataObject =
        this.bpelGenerator.getOutputData(dataOutputAssociations, invokeOperationReference, outAssign, false, bpmnMap);
      if (dataObject instanceof FromParts) {
        invoke.setFromParts((FromParts) dataObject);
      } else if (dataObject instanceof Variable) {
        final Variable outVariable = (Variable) dataObject;
        scope.getVariables().getVariables().add(outVariable);
        invoke.setOutputVariable(outVariable.getName());
      }
    }
    // set toParts
    final List<DataInputAssociation> dataInputAssociations = serviceTask.getDataInputAssociations();
    if (!dataInputAssociations.isEmpty()) {
      final Object dataObject =
        this.bpelGenerator.getInputData(dataInputAssociations, invokeOperationReference, inAssign, true, bpmnMap);
      if (dataObject instanceof ToParts) {
        invoke.setToParts((ToParts) dataObject);
      } else if (dataObject instanceof Variable) {
        final Variable inVariable = (Variable) dataObject;
        scope.getVariables().getVariables().add(inVariable);
        invoke.setInputVariable(inVariable.getName());
      }
    }

    if (inAssign.getCopiesAndExtensionAssignOperations().isEmpty()
        && outAssign.getCopiesAndExtensionAssignOperations().isEmpty()) {
      return invoke;
    } else {
      final Sequence sequence = new Sequence();
      sequence.setName(serviceTask.getId() + "_sequence");
      final Documentation sequenceDoc = new Documentation();
      sequenceDoc.getContent().add("Sequence for " + serviceTask.getName());
      sequence.getDocumentations().add(sequenceDoc);
      if (!inAssign.getCopiesAndExtensionAssignOperations().isEmpty()) {
        sequence.getAssignsAndCompensatesAndCompensateScopes().add(inAssign);
      }
      sequence.getAssignsAndCompensatesAndCompensateScopes().add(invoke);
      if (!outAssign.getCopiesAndExtensionAssignOperations().isEmpty()) {
        sequence.getAssignsAndCompensatesAndCompensateScopes().add(outAssign);
      }
      if (scope.getVariables().getVariables().isEmpty()) {
        return sequence;
      } else {
        scope.setSequence(sequence);
        return scope;
      }
    }
  }

}
