/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.ArrayList;
import java.util.List;
import java.util.ListIterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.EmptyComponent;
import org.ow2.orchestra.bpmn2bpel.components.JoinConditionComponent;
import org.ow2.orchestra.bpmn2bpel.components.LinkComponent;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TExpression;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.jaxb.bpmn.TParallelGateway;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public final class FlowConditionLinkTransformer extends AbstractFlowLinkTransformer {

  private static final Logger LOG = Logger.getLogger(FlowConditionLinkTransformer.class.getName());

  public boolean replaceFlow(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(FlowConditionLinkTransformer.LOG, Level.INFO, "Trying to create link in a flow with a condition");
    // get switch gateway
    for (final TFlowNode node : new ArrayList<TFlowNode>(bpmnMap.getFlowNodes().values())) {
      if (FlowLinkTransformer.isSynchronizingProcessComponent(bpmnMap, node, null, false)) {
        if (node instanceof TExclusiveGateway) {
          final List<TSequenceFlow> incomingSequenceFlows = bpmnMap.getIncomingSequenceFlows(node);
          final List<TSequenceFlow> outgoingSequenceFlows = bpmnMap.getOutgoingSequenceFlows(node);
          if (outgoingSequenceFlows.size() > 1) {
            // node is a exclusive diverging gateway -> remove sequenceFlows and replace with links
            if (this.replaceOutgoingSequenceFlowsWithLink(node, incomingSequenceFlows, outgoingSequenceFlows, bpmnMap)) {
              return true;
            }
          }
        }
      }
    }
    // get merge gateway
    for (final TFlowNode node : new ArrayList<TFlowNode>(bpmnMap.getFlowNodes().values())) {
      if (FlowLinkTransformer.isSynchronizingProcessComponent(bpmnMap, node, null, true)) {
        if (node instanceof TExclusiveGateway) {
          final List<TSequenceFlow> incomingSequenceFlows = bpmnMap.getIncomingSequenceFlows(node);
          final List<TSequenceFlow> outgoingSequenceFlows = bpmnMap.getOutgoingSequenceFlows(node);
          if (incomingSequenceFlows.size() > 1) {
            // node is a exclusive converging gateway -> remove sequenceFlows and replace with links
            if (this.replaceIncomingSequenceFlowsWithLink(node, incomingSequenceFlows, outgoingSequenceFlows, bpmnMap)) {
              return true;
            }
          }
        }
      }
    }
    return false;
  }

  /**
   * @param node
   * @param outgoingSequenceFlows
   * @param bpmnMap
   * @return
   */
  private boolean replaceOutgoingSequenceFlowsWithLink(
          final TFlowNode gateway,
          final List<TSequenceFlow> incomingSequenceFlows,
          final List<TSequenceFlow> sequencesToReplace,
          final BpmnMap bpmnMap) {

    if (incomingSequenceFlows.size() != 1) {
      throw new TransformerException("Bidirectionnal gateways not supported");
    }

    TFlowNode sourceNode = (TFlowNode) incomingSequenceFlows.get(0).getSourceRef();
    if (sourceNode instanceof TGateway) {
      // don't create link to a gateway, insert empty activity
      final EmptyComponent emptyComponent = new EmptyComponent();
      emptyComponent.setId(Misc.getHumanReadableId("component"));
      bpmnMap.getFlowNodes().put(new QName(emptyComponent.getId()), emptyComponent);

      // new sequence flow starts from empty component, so as to be sure the condition from the gateway is not lost
      bpmnMap.changeSequenceFlowTarget(incomingSequenceFlows.get(0), emptyComponent);
      bpmnMap.createSequenceFlow(emptyComponent, gateway);

      Misc.fastDynamicLog(FlowConditionLinkTransformer.LOG, Level.INFO,
              "Adding empty activity %s before decision gateway: %s (%s)",
              emptyComponent.getId(), gateway.getId(), gateway);
      // use emptyActivity as sourceNode
      sourceNode = emptyComponent;
    }

    // sourceNode is an Activity or an event
    for (final TSequenceFlow sequenceToReplace : sequencesToReplace) {
      final TFlowNode targetNode = (TFlowNode) sequenceToReplace.getTargetRef();

      // Add link
      TFlowNode linkTarget = targetNode;
      // if transition between 2 gateways, use next node or add empty node
      if (targetNode instanceof TGateway) {
        // don't create link to a gateway, insert empty activity
        final EmptyComponent emptyComponent = new EmptyComponent();
        emptyComponent.setId(Misc.getHumanReadableId("component"));
        bpmnMap.getFlowNodes().put(new QName(emptyComponent.getId()), emptyComponent);

        // new sequence flow starts from empty component, so as to be sure the condition from the gateway is not lost
        bpmnMap.changeSequenceFlowTarget(sequenceToReplace, emptyComponent);
        bpmnMap.createSequenceFlow(emptyComponent, targetNode);

        Misc.fastDynamicLog(FlowConditionLinkTransformer.LOG, Level.INFO,
                "Adding empty activity %s after decision gateway: %s (%s)",
                emptyComponent.getId(), gateway.getId(), gateway);
        // use this empty activity as link source.
        linkTarget = emptyComponent;
      }
      final LinkComponent linkComponent = new LinkComponent(
              sourceNode,
              linkTarget,
              this.replaceTransitionCondition(sequenceToReplace, sequencesToReplace));
      bpmnMap.getLinks().add(linkComponent);
      JoinConditionComponent.addNewAndLink(bpmnMap, linkComponent);

      // remove sequenceFlow
      bpmnMap.deleteSequenceFlow(sequenceToReplace);

      // if linkTarget has no more incoming sequenceFlow, add a sequenceFlow link from start
      if (bpmnMap.getIncomingSequenceFlows(linkTarget).isEmpty()) {
        if (bpmnMap.getStartEvents().size() != 1) {
          throw new TransformerException("Only process with one start event is supported.");
        }
        // create a parallel gateway after start node
        final TParallelGateway firstGateway = this.getOrAddStartParallelGateway(bpmnMap);
        // create sequenceFlow from the new gateway to the target of the link.
        bpmnMap.createSequenceFlow(firstGateway, linkTarget);
      }
      Misc.fastDynamicLog(FlowConditionLinkTransformer.LOG, Level.INFO, "Creating link: %s and removing sequenceFlow %s: %s (%s) -> %s (%s)",
              linkComponent.getName(),
              sequenceToReplace.getId(),
              ((TFlowNode) sequenceToReplace.getSourceRef()).getId(),
              sequenceToReplace.getSourceRef().getClass().getSimpleName(),
              ((TFlowNode) sequenceToReplace.getTargetRef()).getId(),
              sequenceToReplace.getTargetRef().getClass().getSimpleName());
    }
    // if sourceNode has no more outgoing sequenceFlow, create a sequenceFlow to endEvent
    if (bpmnMap.getEndEvents().size() != 1) {
      throw new TransformerException("Only process with one end event is supported.");
    }
    // create a parallel gateway before last node
    final TParallelGateway lastGateway = this.getOrAddEndParallelGateway(bpmnMap);

    // create sequenceFlow from the source of the links to the new gateway.
    bpmnMap.createSequenceFlow(gateway, lastGateway);

    return true;
  }

  private boolean replaceIncomingSequenceFlowsWithLink(
          final TFlowNode gateway,
          final List<TSequenceFlow> sequencesToReplace,
          final List<TSequenceFlow> outSequences,
          final BpmnMap bpmnMap) {

    if (outSequences.size() != 1) {
      throw new TransformerException("Bidirectionnal gateways not supported");
    }

    final TSequenceFlow outSequence = outSequences.get(0);
    TFlowNode targetNode = (TFlowNode) outSequence.getTargetRef();
    if (targetNode instanceof TGateway) {
      // don't create link to a gateway, insert empty activity
      final EmptyComponent emptyComponent = new EmptyComponent();
      emptyComponent.setId(Misc.getHumanReadableId("component"));
      bpmnMap.getFlowNodes().put(new QName(emptyComponent.getId()), emptyComponent);

      bpmnMap.changeSequenceFlowSource(outSequence, emptyComponent);
      final TSequenceFlow newSeqFlow = bpmnMap.createSequenceFlow(gateway, emptyComponent);
      // Move condition to created sequenceFlow
      newSeqFlow.setConditionExpression(outSequence.getConditionExpression());
      outSequence.setConditionExpression(null);

      Misc.fastDynamicLog(FlowConditionLinkTransformer.LOG, Level.INFO,
              "Adding empty activity %s after merge gateway: %s (%s)",
              emptyComponent.getId(), gateway.getId(), gateway);
      // use emptyActivity as targetNode
      targetNode = emptyComponent;
    }

    // create a parallel gateway before last node
    final TParallelGateway lastGateway = this.getOrAddEndParallelGateway(bpmnMap);

    // Target is a Activity or an event
    final List<LinkComponent> createdLinkComponents = new ArrayList<LinkComponent>();
    for (final TSequenceFlow sequenceToReplace : sequencesToReplace) {
      final TFlowNode sourceNode = (TFlowNode) sequenceToReplace.getSourceRef();

      // Add link
      TFlowNode linkSource = sourceNode;
      // if transition between 2 gateways, add empty node
      if (sourceNode instanceof TGateway) {
        // don't create link to a gateway, insert empty activity
        final EmptyComponent emptyComponent = new EmptyComponent();
        emptyComponent.setId(Misc.getHumanReadableId("component"));
        bpmnMap.getFlowNodes().put(new QName(emptyComponent.getId()), emptyComponent);

        bpmnMap.changeSequenceFlowSource(sequenceToReplace, emptyComponent);
        final TSequenceFlow newSeqFlow = bpmnMap.createSequenceFlow(sourceNode, emptyComponent);
        // move condition to created sequence
        newSeqFlow.setConditionExpression(sequenceToReplace.getConditionExpression());

        Misc.fastDynamicLog(FlowConditionLinkTransformer.LOG, Level.INFO,
                "Adding empty activity %s before merge gateway: %s (%s)",
                emptyComponent.getId(), gateway.getId(), gateway);
        // use this empty activity as link source.
        linkSource = emptyComponent;
      }
      final LinkComponent linkComponent = new LinkComponent(linkSource, targetNode, null);
      bpmnMap.getLinks().add(linkComponent);
      createdLinkComponents.add(linkComponent);

      // remove sequenceFlow
      bpmnMap.deleteSequenceFlow(sequenceToReplace);

      // if sourceNode has no more outgoing sequenceFlow, add a sequenceFlow link to end
      if (bpmnMap.getEndEvents().size() != 1) {
        throw new TransformerException("Only process with one end event is supported.");
      }

      // create sequenceFlow from the source of the link to the new gateway.
      bpmnMap.createSequenceFlow(linkSource, lastGateway);
      Misc.fastDynamicLog(FlowConditionLinkTransformer.LOG, Level.INFO, "Creating link: %s and removing sequenceFlow %s: %s (%s) -> %s (%s)",
              linkComponent.getName(),
              sequenceToReplace.getId(),
              ((TFlowNode) sequenceToReplace.getSourceRef()).getId(),
              sequenceToReplace.getSourceRef().getClass().getSimpleName(),
              ((TFlowNode) sequenceToReplace.getTargetRef()).getId(),
              sequenceToReplace.getTargetRef().getClass().getSimpleName());
    }
    // create joinCondition
    JoinConditionComponent.addNewAndCondition(bpmnMap, JoinConditionComponent.getOrCondition(createdLinkComponents), targetNode);

    // if targetNode has no more incoming sequenceFlow, create a sequenceFlow from startEvent
    if (bpmnMap.getStartEvents().size() != 1) {
      throw new TransformerException("Only process with one start event is supported.");
    }
    // create a parallel gateway before first node
    final TParallelGateway firstGateway = this.getOrAddStartParallelGateway(bpmnMap);

    // create sequenceFlow from the new gateway to the target of the links.
    bpmnMap.createSequenceFlow(firstGateway, gateway);

    return true;
  }

  private TExpression replaceTransitionCondition(
          final TSequenceFlow sequenceToReplace,
          final List<TSequenceFlow> sequencesToReplace) {
    final TExpression tExpression = new TExpression();
    String expression = "";
    final int sequenceIndex = sequencesToReplace.indexOf(sequenceToReplace);
    // add before branches conditions
    // all branches have conditions, only the last branch can have no conditions
    for (int i = 0; i < sequenceIndex; i++) {
      final TSequenceFlow otherSeqFlow = sequencesToReplace.get(i);
      final ListIterator<Object> iterator = otherSeqFlow.getConditionExpression().getContent().listIterator();
      while (iterator.hasNext()) {
        final Object o = iterator.next();
        if (o instanceof String) {
          final String expr = (String) o;
          expression += "not(" + expr + ")";
          if (i < sequenceIndex - 1) {
            expression += " and ";
          }
        }
      }
    }
    // add this branch condition
    if (sequenceToReplace.getConditionExpression() != null) {
      final ListIterator<Object> iterator = sequenceToReplace.getConditionExpression().getContent().listIterator();
      while (iterator.hasNext()) {
        final Object o = iterator.next();
        if (o instanceof String) {
          final String expr = (String) o;
          expression += expr;
        }
      }
    }
    tExpression.getContent().add(expression);
    return tExpression;
  }
}
