/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Set;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.OnEventComponent;
import org.ow2.orchestra.jaxb.bpmn.TEndEvent;
import org.ow2.orchestra.jaxb.bpmn.TEventBasedGateway;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public final class OnEventTransformer {

  private static final Logger LOG = Logger.getLogger(OnEventTransformer.class.getName());

  /**
   * Get and replace all sequenceFlow by onEvent.
   * Does not replace the start and end event activities
   *
   * @param bpmnMap
   * @return true if something was replaced
   */
  public boolean replaceWithOnEvent(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(OnEventTransformer.LOG, Level.INFO, "Trying to replace with OnEvents");
    TFlowNode startEvent = bpmnMap.getStartEvents().values().iterator().next();
    if (bpmnMap.getFlowNodes().containsValue(startEvent)) {
      TFlowNode nextNode = (TFlowNode) bpmnMap.getOutgoingSequenceFlows(startEvent).get(0).getTargetRef();
      if (!(nextNode instanceof OnEventComponent)) {
        Set<TEndEvent> endEvents = new HashSet<TEndEvent>();
        OnEventComponent newOnEventComponent = replaceNodeWithOnEvent(nextNode, bpmnMap, new HashSet<TFlowNode>(), endEvents);
        cleanSequenceFlows(newOnEventComponent, new HashSet<OnEventComponent>(), bpmnMap);
        bpmnMap.getFlowNodes().put(new QName(newOnEventComponent.getId()), newOnEventComponent);
        for (TEndEvent endEvent : endEvents) {
          bpmnMap.createSequenceFlow(newOnEventComponent, endEvent);
        }
        bpmnMap.createSequenceFlow(startEvent, newOnEventComponent);
        return true;
      }
    }
    return false;
  }

  private void cleanSequenceFlows(final OnEventComponent onEventComponent, Set<OnEventComponent> result, BpmnMap bpmnMap) {
    result.add(onEventComponent);
    for (TSequenceFlow outSequenceFlow : onEventComponent.getOutSequenceFlows()) {
      if ((outSequenceFlow.getTargetRef() instanceof OnEventComponent) && !result.contains(outSequenceFlow.getTargetRef())) {
        cleanSequenceFlows((OnEventComponent) outSequenceFlow.getTargetRef(), result, bpmnMap);
      }
    }
    for (TSequenceFlow seqFlow : bpmnMap.getIncomingSequenceFlows(onEventComponent)) {
      bpmnMap.deleteSequenceFlow(seqFlow);
    }
    for (TSequenceFlow seqFlow : bpmnMap.getOutgoingSequenceFlows(onEventComponent)) {
      bpmnMap.deleteSequenceFlow(seqFlow);
    }
  }

  private OnEventComponent replaceNodeWithOnEvent(TFlowNode node, BpmnMap bpmnMap, Set<TFlowNode> processedNodes, Set<TEndEvent> endEvents) {
    if (processedNodes.contains(node)) {
      return null;
    }
    processedNodes.add(node);
    List<TSequenceFlow> outgoingSequenceFlows = bpmnMap.getOutgoingSequenceFlows(node);
    List<TFlowNode> pickNodes = new ArrayList<TFlowNode>();
    if (node instanceof TEventBasedGateway) {
      List<TSequenceFlow> newOutgoingSequenceFlows = new ArrayList<TSequenceFlow>();
      for (TSequenceFlow outgoingSequenceFlow : outgoingSequenceFlows) {
        // Pick -> store next node and process next next
        TFlowNode pickNode = (TFlowNode) outgoingSequenceFlow.getTargetRef();
        pickNodes.add(pickNode);
        outgoingSequenceFlow = bpmnMap.getOutgoingSequenceFlows(pickNode).get(0);
        newOutgoingSequenceFlows.add(outgoingSequenceFlow);
      }
      outgoingSequenceFlows = newOutgoingSequenceFlows;
    }
    // Create OnEventComponent
    OnEventComponent onEventComponent = new OnEventComponent(node, pickNodes, outgoingSequenceFlows, bpmnMap.getIncomingSequenceFlows(node));
    onEventComponent.setId(Misc.getHumanReadableId("component"));

    Misc.fastDynamicLog(OnEventTransformer.LOG, Level.INFO, "Trying to replace %s (%s) with OnEventComponent (%s)",
            node.getId(),
            node.getClass().getSimpleName(),
            onEventComponent.getId());

    // Get list of next nodes
    for (TSequenceFlow outgoingSequenceFlow : onEventComponent.getOutSequenceFlows()) {
      // Generate onEventComponent for target
      if (outgoingSequenceFlow.getTargetRef() instanceof TEndEvent) {
        endEvents.add((TEndEvent) outgoingSequenceFlow.getTargetRef());
      } else if (!(outgoingSequenceFlow.getTargetRef() instanceof OnEventComponent)) {
        replaceNodeWithOnEvent((TFlowNode) outgoingSequenceFlow.getTargetRef(), bpmnMap, processedNodes, endEvents);
      }
    }

    // replace incoming transitions
    for (TSequenceFlow incomingSequenceFlow : onEventComponent.getInSequenceFlows()) {
      bpmnMap.changeSequenceFlowTarget(incomingSequenceFlow, onEventComponent);
    }
    // replace outgoing transitions
    for (TSequenceFlow outgoingSequenceFlow : onEventComponent.getOutSequenceFlows()) {
      bpmnMap.changeSequenceFlowSource(outgoingSequenceFlow, onEventComponent);
    }
    // remove pickNodes
    for (TFlowNode flowNode : onEventComponent.getPickOutNodes()) {
      bpmnMap.removeNode(flowNode);
    }
    // remove node
    bpmnMap.removeNode(node);
    // return
    return onEventComponent;
  }

}
