/**
 * Copyright (C) 2010  Bull S. A. S.
 * Bull, Rue Jean Jaures, B.P.68, 78340, Les Clayes-sous-Bois
 * This library is free software; you can redistribute it and/or modify it under the terms
 * of the GNU Lesser General Public License as published by the Free Software Foundation
 * version 2.1 of the License.
 * This library is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY;
 * without even the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
 * See the GNU Lesser General Public License for more details.
 * You should have received a copy of the GNU Lesser General Public License along with this
 * program; if not, write to the Free Software Foundation, Inc., 51 Franklin Street, Fifth
 * Floor, Boston, MA  02110-1301, USA.
 **/
package org.ow2.orchestra.bpmn2bpel.transformers;

import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.logging.Level;
import java.util.logging.Logger;
import javax.xml.namespace.QName;
import org.ow2.orchestra.bpmn2bpel.BpmnMap;
import org.ow2.orchestra.bpmn2bpel.components.EmptyComponent;
import org.ow2.orchestra.bpmn2bpel.components.PickComponent;
import org.ow2.orchestra.bpmn2bpel.exceptions.TransformerException;
import org.ow2.orchestra.jaxb.bpmn.TEventBasedGateway;
import org.ow2.orchestra.jaxb.bpmn.TExclusiveGateway;
import org.ow2.orchestra.jaxb.bpmn.TFlowNode;
import org.ow2.orchestra.jaxb.bpmn.TGateway;
import org.ow2.orchestra.jaxb.bpmn.TIntermediateCatchEvent;
import org.ow2.orchestra.jaxb.bpmn.TSequenceFlow;
import org.ow2.orchestra.util.Misc;


/**
 * @author Guillaume Porcher
 */
public final class PickTransformer {

  private static final Logger LOG = Logger.getLogger(PickTransformer.class.getName());

  /**
   * Get and replace a flow
   *
   * @param bpmnMap
   * @return
   */
  public boolean replacePick(final BpmnMap bpmnMap) {
    Misc.fastDynamicLog(PickTransformer.LOG, Level.INFO, "Trying to replace pick");
    final List<TFlowNode> nodes = this.getFirstPickComponent(bpmnMap);
    if (nodes.size() < 3) {
      return false;
    }
    final TFlowNode firstEvent = nodes.get(0);
    final TFlowNode firstActivity = nodes.get(1);


    // start gateway is the source of the first node incoming sequenceFlow
    final TEventBasedGateway startGateway =
            (TEventBasedGateway) bpmnMap.getIncomingSequenceFlows(firstEvent).iterator().next().getSourceRef();
    // endGateway id the target of the first node outgoing sequenceFlow
    final TExclusiveGateway endGateway =
            (TExclusiveGateway) bpmnMap.getOutgoingSequenceFlows(firstActivity).iterator().next().getTargetRef();

    // flowNodes is a valid pick, replace it with a component
    final PickComponent pickComponent = new PickComponent(nodes);
    pickComponent.setId(Misc.getHumanReadableId("component"));
    bpmnMap.getFlowNodes().put(new QName(pickComponent.getId()), pickComponent);

    // remove nodes
    // Replace first sequenceFlow
    bpmnMap.createSequenceFlow(startGateway, pickComponent);
    // replace last sequence flow
    bpmnMap.createSequenceFlow(pickComponent, endGateway);
    // remove flow nodes from list
    for (final TFlowNode node : nodes) {
      bpmnMap.removeNode(node);
    }

    // check if gateways can be removed is done in a separate transformer.
    Misc.fastDynamicLog(PickTransformer.LOG, Level.INFO, "Creating pick component: %s", pickComponent);
    return true;
  }

  private List<TFlowNode> getFirstPickComponent(final BpmnMap bpmnMap) {
    List<TFlowNode> flowNodes = Collections.emptyList();
    for (final TFlowNode flowNode : new ArrayList<TFlowNode>(bpmnMap.getFlowNodes().values())) {
      if (flowNode instanceof TEventBasedGateway) {
        flowNodes = this.getFirstPickComponent(bpmnMap, flowNode);
        if (!flowNodes.isEmpty()) {
          break;
        }
      }
    }
    return flowNodes;
  }

  private List<TFlowNode> getFirstPickComponent(final BpmnMap bpmnMap, final TFlowNode startNode) {
    if (!(startNode instanceof TEventBasedGateway)) {
      return Collections.emptyList();
    }
    final List<TFlowNode> nodes = new ArrayList<TFlowNode>();
    TExclusiveGateway pickEndGateway = null;
    final Collection<TSequenceFlow> outgoings = bpmnMap.getOutgoingSequenceFlows(startNode);
    for (final TSequenceFlow seq : outgoings) {
      final TFlowNode target = (TFlowNode) seq.getTargetRef();
      if (!(target instanceof TIntermediateCatchEvent)) {
        throw new TransformerException(target.getClass().getSimpleName() + " not allowed after an event based gateway.");
      }
      final TIntermediateCatchEvent event = (TIntermediateCatchEvent) target;
      // next activity can be an activity (followed by a TExclusiveGateway) or directly a TExclusiveGateWay
      // all branches of the pick should join the same gateway.
      final Collection<TSequenceFlow> targetOut = bpmnMap.getOutgoingSequenceFlows(target);
      if (targetOut.size() == 1) {
        final TSequenceFlow targetSeq = targetOut.iterator().next();
        TFlowNode eventActivity = (TFlowNode) targetSeq.getTargetRef();
        if (eventActivity instanceof TExclusiveGateway) {
          // direct connection to a gateway, add an emptyComponent
          final EmptyComponent emptyComponent = new EmptyComponent();
          emptyComponent.setId(Misc.getHumanReadableId("component"));
          bpmnMap.getFlowNodes().put(new QName(emptyComponent.getId()), emptyComponent);

          // new sequence flow starts from empty component, so as to be sure the condition from the gateway is not lost
          bpmnMap.changeSequenceFlowTarget(targetSeq, emptyComponent);
          bpmnMap.createSequenceFlow(emptyComponent, eventActivity);
          Misc.fastDynamicLog(PickTransformer.LOG, Level.INFO,
                  "Adding empty activity %s before merge gateway: %s (%s)",
                  emptyComponent.getId(), eventActivity.getId(), eventActivity);
          // use emptyActivity as eventActivity
          eventActivity = emptyComponent;
        }
        if (eventActivity instanceof TGateway) {
          return Collections.emptyList();
        }
        // get nextActivity next activity
        final List<TSequenceFlow> nextActivityOut = bpmnMap.getOutgoingSequenceFlows(eventActivity);
        if (nextActivityOut.size() != 1) {
          return Collections.emptyList();
        }
        final Object activityGateway = nextActivityOut.get(0).getTargetRef();
        if (activityGateway instanceof TExclusiveGateway) {
          if (pickEndGateway == null) {
            pickEndGateway = (TExclusiveGateway) activityGateway;
          } else if (!pickEndGateway.equals(activityGateway)) {
            return Collections.emptyList();
          }
          nodes.add(event);
          nodes.add(eventActivity);
        } else {
          return Collections.emptyList();
        }
      }
    }
    return nodes;
  }

}
